#include "part1-corrige.hh"


// ==== Part 1, question 1 code hereafter

// Q-111
ref make_ref(const std::string& value) {
  return std::make_shared<std::string>(value);
}

// Q-112 : handle the nice printing of , 
std::ostream& operator<<(std::ostream& os, const ref_set& rs) {
  os << '{';
  auto it = rs.begin();
  if(it != rs.end()) {
    os << *(*(it++));
    while(it != rs.end())
      os << ", " << *(*(it++));
  }
  os << '}';

  return os;
}


// ==== Part 1, question 2 code hereafter

// Q-121 : Default constructor for the result is ok. Do not return
// make_ref(""), it would be a non empty set, having one element which
// is the empty string... it would display as {} as well, but this is
// wrong.
ref_set empty_set() {return {};}

// Q-122 : compare pointed values (*e == *elem), not pointers. Indeed,
// if two pointers point on two instances of the same string,
// (e==elem) is false, while the pointed values are the same string.
bool has(const ref_set& the_set, ref elem) {
  for(auto e : the_set)
    if(*e == *elem)
      return true;
  return false;
}

// ==== Part 1, question 3 code hereafter

// Q-131 : Insert an element if it is not already there.
void insert(ref_set& the_set, ref elem) {
  if(!has(the_set, elem))
    the_set.push_back(elem);
}

// Q-132 : Just a call to insert.
void operator+=(ref_set& the_set, ref elem) {
  insert(the_set, elem);
}

// Q-133 : With the above +=, we take apart in tmp the whole data of
// the_set (std::move), the_set becomes an empty collection. Then we
// add them back in the_set, avoiding duplicates since += checks this.
void remove_duplicates(ref_set& the_set) {
  auto tmp = std::move(the_set);
  for(auto e : tmp) the_set += e;
}


// ==== Part 1, question 4 code hereafter

// Q-141 : The union is at least A (res is iniialized as a copy of A),
// where elements of B are inserted if needed (+= checks this).
ref_set operator||(const ref_set& A, const ref_set& B) {
  ref_set res = A;
  for(auto e : B) res += e;
  return res;
}

// Q-142 : We start from an empty res, and inserts the elements of A
// that also belong to B.
ref_set operator&&(const ref_set& A, const ref_set& B) {
  ref_set res;
  for(auto e : A)
    if(has(B, e)) res += e;
  return res;
}

// Q-143 : If both sises are identical, we guess that there are no
// duplicated, and we check that every element of A can be found in
// B. This is robust to different ordering of the same elements in A
// and B.
bool operator==(const ref_set& A, const ref_set& B) {
  if(A.size() != B.size()) return false;
  for(auto e : A)
    if(!has(B, e)) return false;
  return true;
}
