
#include <iostream>
#include <string>
#include <memory>
#include <vector>
#include <initializer_list>

#include <functional>

// Q31 : If you have inlined the code in part 2 (as we did) and if you
// have used content_type instead of std::string everywhere in part 2
// (as we dis as well), you only need two slight modification to make
// a template from the code of part 2:
template<typename CONTENT>                                // One here
class ref_set {
public:
  using content_type = CONTENT;                           // The second here... and that's it !
  using ref_type     = std::shared_ptr<content_type>;

private:

  std::vector<ref_type> data;

  friend std::ostream& operator<<(std::ostream& os, const ref_set& rs) {
    os << '{';
    auto it = rs.data.begin();
    if(it != rs.data.end()) {
      os << *(*(it++));
      while(it != rs.data.end())
	os << ", " << *(*(it++));
    }
    os << '}';

    return os;
  }


  void remove_duplicates() {
    auto tmp = std::move(data);
    for(auto e : tmp) (*this) += e;
  }
  
public:

  ref_set()                          = default;
  ref_set(const ref_set&)            = default;
  ref_set(ref_set&&)                 = default;
  ref_set& operator=(const ref_set&) = default;
  ref_set& operator=(ref_set&&)      = default;

  ref_set(const std::initializer_list<ref_type>& data) : data(data) {remove_duplicates();}

  static ref_type make_ref(const content_type& value) {return std::make_shared<content_type>(value);}
  static ref_set empty_set() {return {};}
  
  bool has(ref_type elem) const {
    for(auto e : data) if(*e == *elem) return true;
    return false;
  }
  
  void operator+=(ref_type elem) {if(!has(elem)) data.push_back(elem);}
  
  ref_set operator||(const ref_set& B) const {
    ref_set res = *this;
    for(auto e : B.data) res += e;
    return res;
  }
  
  ref_set operator&&(const ref_set& B) const {
    ref_set res;
    for(auto e : data)
      if(B.has(e)) res += e;
    return res;
  }

  bool operator==(const ref_set& B) const {
    if(data.size() != B.data.size()) return false;
    for(auto e : data)
      if(!B.has(e)) return false;
    return true;
  }
  
  ref_set apply(const std::function<content_type (const content_type)>& f) {
    ref_set res;
    for(auto e : data)
      res += make_ref(f(*e));
    return res;
  }

  template<typename OTHER_CONTENT> friend class ref_set; // This is required by Q32

  // Q32 : This is very difficult. You need to allow your class to
  // handle a ref_set of another type, this is why all the ref_set<T>
  // should be friend (we have added a friend statement for this just
  // above. Then, the problem is to handle the recursion...
  ref_set<ref_set<content_type>> power_set() const {
    // It this = {a, b, c}
    if(data.size() == 0)
      return {ref_set<ref_set<content_type>>::make_ref(empty_set())};
    ref_set tail;
    for(auto it = data.begin() + 1; it != data.end(); ++it) tail += *it;
    // tail = {b, c}
    ref_set head = {*(data.begin())};
    // head = {a}
    auto power_tail = tail.power_set();
    // power_tail = {{}, {b}, {c}, {b, c}}
    ref_set<ref_set<content_type>> power_head;
    for(auto e : power_tail.data) // ex : *e = {b, c} 
      power_head += ref_set<ref_set<content_type>>::make_ref(*e || head);
    return power_tail || power_head;
  }
};
